#include "winmem.h"

#include <plugin.hpp>

#include "collect.h"
#include "find-all.h"
#include "fall-lng.h"
#include "cregexp.h"
#include "scanreg.h"
#include "FARIntf.h"

#ifdef DEBUG
#include <stdio.h>
static void SysLog(char *fmt,...)
{
  va_list argptr;
  va_start(argptr, fmt);
  FILE *LogStream = fopen("C:\\BC5\\WORK\\PLUGINS\\FIND-ALL\\FIND-ALL.LOG", "at");
  vfprintf(LogStream, fmt, argptr);
  fflush(LogStream);
  fclose(LogStream);
  va_end(argptr);
}
#else
#define SysLog //
#endif

static char SavedPattern[512];

static struct TOptions
{
  int PickWord,ZeroWarn,Case,RegExp,Lines,Global,Before,After,CFunc,CRegNum;
  char CMatch[256],CIgnore[256];
  void GetCFunc(char*, char*);
  void Customize(char*);
  void Get(void);
  void Put(void);
  void customizeLanguage(char*);
} Opt = { 1,1,0,0,0,1,0,0,0,1,"","" };

void TOptions::Get(void)
{
  PickWord = GetRegKey(HKEY_CURRENT_USER,"","PickWord",1);
  ZeroWarn = GetRegKey(HKEY_CURRENT_USER,"","ZeroWarn",1);
  Case     = GetRegKey(HKEY_CURRENT_USER,"","Case",    0);
  RegExp   = GetRegKey(HKEY_CURRENT_USER,"","RegExp",  0);
  Lines    = GetRegKey(HKEY_CURRENT_USER,"","Lines",   0);
  Global   = GetRegKey(HKEY_CURRENT_USER,"","Global",  1);
  Before   = GetRegKey(HKEY_CURRENT_USER,"","Before",  0);
  After    = GetRegKey(HKEY_CURRENT_USER,"","After",   0);
  CFunc    = GetRegKey(HKEY_CURRENT_USER,"","CFunc",   0);
}

void TOptions::Put(void)
{
  SetRegKey(HKEY_CURRENT_USER, "", "PickWord", PickWord);
  SetRegKey(HKEY_CURRENT_USER, "", "ZeroWarn", ZeroWarn);
  SetRegKey(HKEY_CURRENT_USER, "", "Case",     Case    );
  SetRegKey(HKEY_CURRENT_USER, "", "RegExp",   RegExp  );
  SetRegKey(HKEY_CURRENT_USER, "", "Lines",    Lines   );
  SetRegKey(HKEY_CURRENT_USER, "", "Global",   Global  );
  SetRegKey(HKEY_CURRENT_USER, "", "Before",   Before  );
  SetRegKey(HKEY_CURRENT_USER, "", "After",    After   );
  SetRegKey(HKEY_CURRENT_USER, "", "CFunc",    CFunc   );
}

static inline int isCharSpace(char c)
{
  return strchr(" \t\r\n", c) != NULL;
}

static char *GetCommaWord(char *Src, char *Word)
{
  if ( !*Src )
    return NULL;
  int WordPos, SkipBrackets = FALSE;
  for ( WordPos = 0 ; *Src ; Src++, WordPos++ )
  {
    if ( *Src == '[' && strchr(Src+1,']') )
      SkipBrackets = TRUE;
    if ( *Src == ']' )
      SkipBrackets = FALSE;
    if ( *Src==',' && !SkipBrackets )
    {
      Word[WordPos] = 0;
      Src++;
      while ( isCharSpace(*Src) )
        Src++;
      return Src;
    }
    else
      Word[WordPos] = *Src;
  }
  Word[WordPos] = 0;
  return Src;
}

static int matched(char *mask, char *value)
{
  char p[80], *m = mask;
  while ( ( m = GetCommaWord(m, p) ) != NULL )
    if ( Info.CmpName(p, value, TRUE) )
      return 1;
  return 0;
}

static char *lTrim(char *s)
{
  char * p = s;
  while ( isspace(*p) )
    p++;
  return strcpy(s, p);
}

static char *rTrim(char *s)
{
  char *p = strchr(s, 0)-1;
  if ( isspace(*p) )
  {
    while ( isspace(*--p) || !*p )
      ;
    *++p = 0;
  }
  return s;
}

static inline char *allTrim(char *s)
{
  return lTrim(rTrim(s));
}

static int str2int(const char *s)
{
  char buff[64];
  int res = 0, sign = 1;
  if ( *allTrim(strcpy(buff, s)) == '-' )
  {
    *buff = '0';
    sign = -1;
  }
  for ( char *p = buff ; *p ; p++ )
  {
    if ( strchr("0123456789", *p) == NULL )
      break;
    res = res*10+(*p-'0');
  }
  return res*sign;
}

void TOptions::customizeLanguage(char* aLang)
{
  char lang[256] = "", tMask[256] = "", tMatch[256] = "", tIgnore[256] = "";
  char rNum[8], FullKeyName[512];
  int tRegNum = 1;
  if ( aLang )
  {
    wsprintf(FullKeyName,"Extensions\\%s", strcpy(lang, aLang));
    tRegNum = GetRegKey(HKEY_CURRENT_USER,FullKeyName,"RegNum", 1);
    GetRegKey(HKEY_CURRENT_USER,FullKeyName,"Mask",   tMask,   "", 255);
    GetRegKey(HKEY_CURRENT_USER,FullKeyName,"Match",  tMatch,  "", 255);
    GetRegKey(HKEY_CURRENT_USER,FullKeyName,"Ignore", tIgnore, "", 255);
  }
  char *M = "FindAll-FuncMatch";
  char *I = "FindAll-FuncIgnore";
  wsprintf(rNum, "%d", tRegNum);
  struct InitDialogItem InitItems[] =
  {
    { DI_DOUBLEBOX,   3, 1,72,  9,0,       0,              (char*)MCTitle   }, // 0
    { DI_TEXT,        5, 2,29,  0,0,       0,              (char*)MLang     }, // 1
    { DI_EDIT,       30, 2,70,  0,0,       0,              lang             }, // 2
    { DI_TEXT,        5, 3,29,  0,0,       0,              (char*)MMask     }, // 3
    { DI_EDIT,       30, 3,70,  0,0,       0,              tMask            }, // 4
    { DI_TEXT,        5, 4,29,  0,0,       0,              (char*)MMatch    }, // 5
    { DI_EDIT,       30, 4,70,  0,(DWORD)M,DIF_HISTORY,    tMatch           }, // 6
    { DI_TEXT,        5, 5,29,  0,0,       0,              (char*)MIgnore   }, // 7
    { DI_EDIT,       30, 5,70,  0,(DWORD)I,DIF_HISTORY,    tIgnore          }, // 8
    { DI_TEXT,        5, 6,29,  0,0,       0,              (char*)MRegNum   }, // 9
    { DI_FIXEDIT,    30, 6,31,  0,0,       0,              rNum             }, //10
    { DI_TEXT,        5, 7, 0,255,0,       0,              NULL             }, //11
    { DI_BUTTON,      0, 8, 0,  0,0,       DIF_CENTERGROUP,(char*)MOK       }, //12
    { DI_BUTTON,      0, 8, 0,  0,0,       DIF_CENTERGROUP,(char*)MCancel   }  //13
  };
  int ExitCode = ExecDialog(InitItems,sizeof(InitItems)/sizeof(InitItems[0]),2,12, NULL);
  if ( ExitCode == 12 )
  {
    if ( !*lang )
      return;
    wsprintf(FullKeyName,"Extensions\\%s", lang);
    tRegNum = str2int(rNum);
    SetRegKey(HKEY_CURRENT_USER, FullKeyName, "Mask",   tMask  );
    SetRegKey(HKEY_CURRENT_USER, FullKeyName, "Match",  tMatch );
    SetRegKey(HKEY_CURRENT_USER, FullKeyName, "Ignore", tIgnore);
    SetRegKey(HKEY_CURRENT_USER, FullKeyName, "RegNum", tRegNum);
  }
}

static bool findMask(HKEY, char *Key, char *key, void *data)
{
  char pattern[256], FullKeyName[512];
  wsprintf(FullKeyName,"%s\\%s", Key, key);
  GetRegKey(HKEY_CURRENT_USER,FullKeyName,"Mask", pattern, "", 255);
  return matched(pattern, (char*)data);
}

void TOptions::GetCFunc(char *fn, char *desc)
{
  char key[512], FullKeyName[512];
  if ( desc )
    *desc = 0;
  CRegNum = *CMatch = *CIgnore = 0;
  if ( FindFirstRegKey(HKEY_CURRENT_USER, "Extensions", findMask, fn, key, 255) )
  {
    wsprintf(FullKeyName,"Extensions\\%s", key);
    if ( desc )
      strcpy(desc, key);
    CRegNum = GetRegKey(HKEY_CURRENT_USER,FullKeyName,"RegNum", 1);
    GetRegKey(HKEY_CURRENT_USER,FullKeyName,"Match",  CMatch,  "", 255);
    GetRegKey(HKEY_CURRENT_USER,FullKeyName,"Ignore", CIgnore, "", 255);
  }
  if ( desc && !*desc )
    strcpy(desc, GetMsg(MCTitle));
}

static bool addMask(HKEY, char*, char *key, FarMenuItem *menu, int *n, void *data)
{
  strcpy(menu->Text, key);
  return (data != NULL && *n == -1) ? strcmp(key, (char*)data) == 0 : false;
}

void TOptions::Customize(char *desc)
{
  FarMenuItem *menu = NULL;
  int Code, n = -1, BreakKeys[5] = { VK_RETURN, VK_INSERT, VK_DELETE, VK_F4, 0 };
  char key[256];
  do
  {
    int Count = MenuFromRegKey(HKEY_CURRENT_USER, "Extensions", addMask, &menu, &n, desc);
    if ( n >= 0 && n < Count )
      menu[n].Selected = true;
    n = Info.Menu(Info.ModuleNumber, -1, -1, 0, FMENU_WRAPMODE,
                  GetMsg(MCTitle), GetMsg(MCBottom),
                  "Extensions", BreakKeys, &Code, menu, Count);
    if ( n >= 0 && n < Count )
    {
      wsprintf(key, "Extensions\\%s", menu[n].Text);
      switch ( Code )
      {
        case 1:
          customizeLanguage(NULL);
          break;
        case 2:
          DeleteRegKey(HKEY_CURRENT_USER, key);
          break;
        case 3:
          customizeLanguage(menu[n].Text);
          break;
        default:
          if ( desc )
            strcpy(desc, menu[n].Text);
          CRegNum = GetRegKey(HKEY_CURRENT_USER,key,"RegNum", 1);
          GetRegKey(HKEY_CURRENT_USER,key,"Match",  CMatch,  "", 255);
          GetRegKey(HKEY_CURRENT_USER,key,"Ignore", CIgnore, "", 255);
          n = -1;
      }
    }
    if ( menu )
      winDel(menu);
    menu = NULL;
  } while ( n != -1 );
}

void WINAPI _export SetStartupInfo(struct PluginStartupInfo *Info)
{
  ::Info = *Info;
  wsprintf(PluginRootKey,"%s\\FIND-ALL",Info->RootKey);
  Opt.Get();
}

void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  static char *PluginMenuStrings[1];
  Info->StructSize = sizeof(*Info);
  Info->Flags = PF_EDITOR|PF_DISABLEPANELS;
  Info->DiskMenuStringsNumber = 0;
  Info->PluginConfigStringsNumber = 0;
  PluginMenuStrings[0] = GetMsg(MMenuString);
  Info->PluginMenuStrings = PluginMenuStrings;
  Info->PluginMenuStringsNumber = 1;
  static char *PluginCfgStrings[1];
  PluginCfgStrings[0]=GetMsg(MConfigMenu);
  Info->PluginConfigStrings=PluginCfgStrings;
  Info->PluginConfigStringsNumber=sizeof(PluginCfgStrings)/sizeof(PluginCfgStrings[0]);
}

static int regErrorShown = FALSE;

static int strMatch(int useRegexp, char *s, char *pattern, int& col, int n, char *res, char *prefix, char *suffix)
{
  if ( useRegexp )
  {
    int ret = FALSE;
    char slashPattern[320];
    static CRegExp reg;
    SMatches m;
    strcat(strcat(strcpy(slashPattern, prefix), pattern), suffix);
    if ( reg.SetExpr(slashPattern) )
    {
      ret = reg.Parse(s, &m) ? TRUE : FALSE;
      col = -1;
      if ( ret )
      {
        col = m.s[0];
        if ( res )
        {
          size_t resLen = (size_t)(m.e[n]-m.s[n]);
          if ( resLen )
            strncpy(res, s+m.s[n], resLen);
          res[resLen] = 0;
        }
      }
    }
    else if ( !regErrorShown )
    {
      char *MsgItems[]={GetMsg(MError),GetMsg(MRegError),slashPattern,GetMsg(MOK)};
      Info.Message(Info.ModuleNumber,FMSG_WARNING,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),1);
      regErrorShown = TRUE;
    }
    return ret;
  }
  else
  {
    char *p = strstr(s, pattern);
    if ( p )
      col = p-s;
    return p ? TRUE : FALSE;
  }
}

static char *StrUpper(char *str)
{
  for ( char *c = str ; *c ; c++ )
  {
    char Ansi, ReverseOem;
    OemToCharBuff(c, &Ansi, 1);
    CharToOemBuff(&Ansi, &ReverseOem, 1);
    if ( IsCharAlpha(Ansi) && ReverseOem == *c )
    {
      *c = (char)CharUpper((LPTSTR)(unsigned char)Ansi);
      CharToOemBuff(c, c, 1);
    }
  }
  return str;
}

static int CheckForEsc()
{
  int ExitCode = FALSE;
  for ( ; ; )
  {
    INPUT_RECORD rec;
    static HANDLE hConInp = GetStdHandle(STD_INPUT_HANDLE);
    DWORD ReadCount;
    PeekConsoleInput(hConInp,&rec,1,&ReadCount);
    if ( !ReadCount )
      break;
    ReadConsoleInput(hConInp,&rec,1,&ReadCount);
    if ( rec.EventType == KEY_EVENT && rec.Event.KeyEvent.wVirtualKeyCode == VK_ESCAPE )
      ExitCode = TRUE;
  }
  return ExitCode;
}

struct TEditorPos
{
  TEditorPos() { Default(); }
  int Row, Col;
  int TopRow,LeftCol;
  void Default(void) { Row = Col = TopRow = LeftCol = -1; }
};

static TEditorPos EditorGetPos(void)
{
  TEditorPos r;
  EditorInfo ei;
  Info.EditorControl(ECTL_GETINFO, &ei);
  r.Row = ei.CurLine;
  r.Col = ei.CurPos;
  r.TopRow = ei.TopScreenLine;
  r.LeftCol = ei.LeftPos;
  return r;
}

static void EditorSetPos(TEditorPos pos)
{
  EditorSetPosition sp;
  sp.CurLine = pos.Row;
  sp.CurPos = pos.Col;
  sp.TopScreenLine = pos.TopRow;
  sp.LeftPos = pos.LeftCol;
  sp.CurTabPos = sp.Overtype = -1;
  Info.EditorControl(ECTL_SETPOSITION,&sp);
}

static void EditorProcessVKey(WORD vkey, int state)
{
  INPUT_RECORD tr;
  tr.EventType = KEY_EVENT;
  tr.Event.KeyEvent.bKeyDown = TRUE;
  tr.Event.KeyEvent.wRepeatCount = 1;
  tr.Event.KeyEvent.wVirtualKeyCode = vkey;
  tr.Event.KeyEvent.dwControlKeyState = state;
  Info.EditorControl(ECTL_PROCESSINPUT, &tr);
}

static void EditorGetString(EditorGetString *gs, int line = -1)
{
  gs->StringNumber = line;
  Info.EditorControl(ECTL_GETSTRING, gs);
}

static int funcMatch(char *s, char *res)
{
  int foundCol, ret = 0;
  if ( *Opt.CIgnore )
  {
    if ( !strMatch(1, s, Opt.CIgnore, foundCol, 0, NULL, "", "") )
      if ( *Opt.CMatch && strMatch(1, s, Opt.CMatch, foundCol, Opt.CRegNum, res, "", "") )
        ret = 1;
  }
  else if ( *Opt.CMatch && strMatch(1, s, Opt.CMatch, foundCol, Opt.CRegNum, res, "", "") )
    ret = 1;
  return ret;
}

static void TruncFuncName(char *s)
{
  if ( strlen(allTrim(s)) > 32 )
  {
    s[30] = 0;
    strcat(s, "...");
  }
}

HANDLE WINAPI _export OpenPlugin(int, int)
{
  static char *h = "SearchText";
  char desc[256], data[512];
  struct EditorGetString gs;
  struct EditorConvertText ect;
  struct EditorInfo ei;
  Info.EditorControl(ECTL_GETINFO, &ei);
  TEditorPos p, oldPos = EditorGetPos();
  if ( Opt.PickWord )
  {
    EditorProcessVKey(VK_RIGHT, LEFT_CTRL_PRESSED);
    TEditorPos p1 = EditorGetPos();
    EditorProcessVKey(VK_LEFT, LEFT_CTRL_PRESSED);
    p = EditorGetPos();
    EditorGetString(&gs);
    if ( gs.StringLength && ( p.Row == p1.Row ) && ( p1.Col > p.Col ) )
    {
      strncpy(data, gs.StringText+p.Col, p1.Col-p.Col);
      data[p1.Col-p.Col] = 0;
      ect.Text = allTrim(data);
      ect.TextLength = strlen(data);
      Info.EditorControl(ECTL_EDITORTOOEM, &ect);
    }
    EditorSetPos(oldPos);
  }
  else
    strcpy(data,SavedPattern);
  Opt.GetCFunc(ei.FileName, desc);
  struct InitDialogItem InitItems[] =
  {
    { DI_DOUBLEBOX,   3, 1,72, 12,0,         0,                             (char*)MTitle    }, // 0
    { DI_TEXT,        5, 2,29,  0,0,         0,                             (char*)MFind     }, // 1
    { DI_EDIT,        5, 3,70,  0,(DWORD)h,  DIF_HISTORY|DIF_USELASTHISTORY,data             }, // 2
    { DI_TEXT,        5, 4, 0,255,0,         0,                             NULL             }, // 3
    { DI_CHECKBOX,    5, 5, 0,  0,Opt.Case,  0,                             (char*)MCase     }, // 4
    { DI_CHECKBOX,    5, 6, 0,  0,Opt.RegExp,0,                             (char*)MRegexp   }, // 5
    { DI_CHECKBOX,    5, 7, 0,  0,Opt.Lines, 0,                             (char*)MLines    }, // 6
    { DI_RADIOBUTTON,40, 5, 0,  0,Opt.Global,DIF_GROUP,                     (char*)MGlobal   }, // 7
    { DI_RADIOBUTTON,40, 6, 0,  0,Opt.Before,0,                             (char*)MBefore   }, // 8
    { DI_RADIOBUTTON,40, 7, 0,  0,Opt.After, 0,                             (char*)MAfter    }, // 9
    { DI_TEXT,        5, 8, 0,255,0,         0,                             NULL             }, //10
    { DI_CHECKBOX,    5, 9, 0,  0,Opt.CFunc, 0,                             (char*)MCFunc    }, //11
    { DI_BUTTON,     40, 9,68,  0,0,         0,                             desc             }, //12
    { DI_TEXT,        5,10, 0,255,0,         0,                             NULL             }, //13
    { DI_BUTTON,      0,11, 0,  0,0,         DIF_CENTERGROUP,               (char*)MOK       }, //14
    { DI_BUTTON,      0,11, 0,  0,0,         DIF_CENTERGROUP,               (char*)MCancel   }  //15
  };
  for ( ; ; )
  {
    int ExitCode = ExecDialog(InitItems,sizeof(InitItems)/sizeof(InitItems[0]),2,14,"interface");
    if ( ExitCode == 12 )
    {
      Opt.Customize(desc);
      continue;
    }
    else if ( ExitCode == 14 )
    {
      char top[256], bottom[256], pattern[512], LastFuncName[512];
      size_t MaxFuncLen = *LastFuncName = 0;
      struct EditorInfo ei;
      Info.EditorControl(ECTL_GETINFO, &ei);
      int lines  = ei.TotalLines;
      Opt.Case   = InitItems[4].Selected;
      Opt.RegExp = InitItems[5].Selected;
      Opt.Lines  = InitItems[6].Selected;
      Opt.Global = InitItems[7].Selected;
      Opt.Before = InitItems[8].Selected;
      Opt.After  = InitItems[9].Selected;
      Opt.CFunc  = InitItems[11].Selected;
      if ( !*strcpy(pattern,data) )
        return INVALID_HANDLE_VALUE;
      strcpy(SavedPattern,data);
      if ( !Opt.Case )
        StrUpper(pattern);
      p.Default();
      int nLen = 1;
      int nStart = Opt.After  ? oldPos.Row : 0;
      int nEnd   = Opt.Before ? oldPos.Row : lines;
      int step   = (nEnd-nStart)/100;
      HANDLE hScreen;
      if ( step > 10 )
        hScreen = Info.SaveScreen(0,0,-1,-1);
      collection strList(10, 5);
      regErrorShown = FALSE;
      for ( p.Row = nStart ; p.Row < nEnd ; p.Row++ )
      {
        if ( ( step > 10 ) && ( (p.Row-nStart) % step == 0 ) )
        {
          if ( CheckForEsc() )
            break;
          char InfoLine[80], FoundLine[80];
          if ( strList.getCount() )
            wsprintf(FoundLine,GetMsg(MFound),strList.getCount());
          else
            *FoundLine = 0;
          wsprintf(InfoLine,GetMsg(MComplete),100*(p.Row-nStart)/(nEnd-nStart));
          char *MsgItems[]={GetMsg(MTitle),InfoLine, FoundLine};
          Info.Message(Info.ModuleNumber,0,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),0);
        }
        EditorSetPos(p);
        EditorGetString(&gs);
        if ( gs.StringLength )
        {
          char *s, *UpperString = NULL, *NewString = winNew(char, gs.StringLength+1);
          memcpy(NewString, gs.StringText, gs.StringLength);
          NewString[gs.StringLength] = 0;
          ect.Text = NewString;
          ect.TextLength = gs.StringLength;
          Info.EditorControl(ECTL_EDITORTOOEM, &ect);
          char *suffix = "/";
          if ( Opt.Case )
            s = NewString;
          else if ( Opt.RegExp )
          {
            suffix = "/i";
            s = NewString;
          }
          else
          {
            UpperString = winNew(char, gs.StringLength+1);
            s = StrUpper(strcpy(UpperString, NewString));
          }
          if ( Opt.CFunc && funcMatch(NewString, LastFuncName) )
            TruncFuncName(LastFuncName);
          int foundCol;
          if ( strMatch(Opt.RegExp, s, pattern, foundCol, 0, NULL, "/", suffix) )
          {
            if ( Opt.Lines )
            {
              char nBuff[16];
              wsprintf(nBuff, "%d", p.Row);
              int nlen = strlen(nBuff);
              if ( nlen > nLen )
                nLen = nlen;
            }
            strList.insert(p.Row, foundCol, NewString,LastFuncName);
            size_t fLen = strlen(LastFuncName);
            if ( MaxFuncLen < fLen )
              MaxFuncLen = fLen;
          }
          if ( UpperString )
            winDel(UpperString);
          if ( NewString )
          winDel(NewString);
        }
      }
      if ( step > 10 )
        Info.RestoreScreen(hScreen);
      EditorSetPos(oldPos);
      size_t count = strList.getCount();
      wsprintf(top,GetMsg(MSelect),data);
      wsprintf(bottom,GetMsg(MTotal),count);
      if ( count )
      {
        struct FarMenuItem *amenu = winNew(FarMenuItem, count);
        if ( amenu )
        {
          for ( size_t i = 0 ; i < count ; i++ )
          {
            char fmt[32], prefixLine[16] = "", prefixFunc[64] = "", foundStr[512];
            TEInfo *te = strList.at(i);
            if ( Opt.Lines )
            {
              wsprintf(fmt, "%%%dd", nLen);
              wsprintf(prefixLine, fmt, te->line+1);
            }
            if ( Opt.CFunc )
            {
              wsprintf(fmt, "%%-%ds", MaxFuncLen);
              wsprintf(prefixFunc, fmt, te->CFunc);
            }
            wsprintf(amenu[i].Text, "%s%s%s", prefixLine, prefixFunc, allTrim(strcpy(foundStr, te->s)));
            amenu[i].Selected = amenu[i].Checked = amenu[i].Separator = FALSE;
          }
          int res = Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_WRAPMODE|FMENU_SHOWAMPERSAND,top,bottom,NULL,NULL,NULL,amenu,count);
          if ( res != -1 )
          {
            p.Default();
            p.Row = strList.at(res)->line;
            p.Col = strList.at(res)->col;
            EditorSetPos(p);
          }
          winDel(amenu);
        }
      }
      else if ( Opt.ZeroWarn )
      {
        char *MsgItems[]={top,bottom,GetMsg(MOK)};
        Info.Message(Info.ModuleNumber,FMSG_WARNING,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),1);
      }
    }
    return INVALID_HANDLE_VALUE;
  }
}
/*
  Find All - Defaults ͻ
  [ ] Pick word under cursor                                         
  [x] Warning when string not found                                  
 Ķ
  [ ] Case sensitive                 () Global search               
  [x] Regular expression             ( ) Before cursor               
  [x] Show line numbers              ( ) After cursor                
 Ķ
  [x] Show function name             [Customize language           ] 
 Ķ
                          [ OK ]  [ Cancel ]                         
 ͼ
*/
int Config(void)
{
  struct InitDialogItem InitItems[] =
  {
    { DI_DOUBLEBOX,   3, 1,72, 12,0,           0,              (char*)MConfig   }, // 0
    { DI_CHECKBOX,    5, 2,70,  0,Opt.PickWord,0,              (char*)MPickWord }, // 1
    { DI_CHECKBOX,    5, 3,70,  0,Opt.ZeroWarn,0,              (char*)MZeroWarn }, // 2
    { DI_TEXT,        5, 4, 0,255,0,           0,              NULL             }, // 3
    { DI_CHECKBOX,    5, 5, 0,  0,Opt.Case,    0,              (char*)MCase     }, // 4
    { DI_CHECKBOX,    5, 6, 0,  0,Opt.RegExp,  0,              (char*)MRegexp   }, // 5
    { DI_CHECKBOX,    5, 7, 0,  0,Opt.Lines,   0,              (char*)MLines    }, // 6
    { DI_RADIOBUTTON,40, 5, 0,  0,Opt.Global,  DIF_GROUP,      (char*)MGlobal   }, // 7
    { DI_RADIOBUTTON,40, 6, 0,  0,Opt.Before,  0,              (char*)MBefore   }, // 8
    { DI_RADIOBUTTON,40, 7, 0,  0,Opt.After,   0,              (char*)MAfter    }, // 9
    { DI_TEXT,        5, 8, 0,255,0,           0,              NULL             }, //10
    { DI_CHECKBOX,    5, 9, 0,  0,Opt.CFunc,   0,              (char*)MCFunc    }, //11
    { DI_BUTTON,     40, 9,68,  0,0,           0,              (char*)MCTitle   }, //12
    { DI_TEXT,        5,10, 0,255,0,           0,              NULL             }, //13
    { DI_BUTTON,      0,11, 0,  0,0,           DIF_CENTERGROUP,(char*)MOK       }, //14
    { DI_BUTTON,      0,11, 0,  0,0,           DIF_CENTERGROUP,(char*)MCancel   }  //15
  };
  for ( ; ; )
  {
    int ExitCode = ExecDialog(InitItems,sizeof(InitItems)/sizeof(InitItems[0]),1,14,"Config");
    if ( ExitCode == 12 )
      Opt.Customize(NULL);
    else if ( ExitCode != 14 )
      return FALSE;
    else
      break;
  }
  Opt.PickWord = InitItems[1].Selected;
  Opt.ZeroWarn = InitItems[2].Selected;
  Opt.Case     = InitItems[4].Selected;
  Opt.RegExp   = InitItems[5].Selected;
  Opt.Lines    = InitItems[6].Selected;
  Opt.Global   = InitItems[7].Selected;
  Opt.Before   = InitItems[8].Selected;
  Opt.After    = InitItems[9].Selected;
  Opt.CFunc    = InitItems[11].Selected;
  Opt.Put();
  return TRUE;
}

int WINAPI _export Configure(int ItemNumber)
{
  if ( ItemNumber )
    return FALSE;
  return Config();
}
